/* 
 * dietsniff - a slim and static network sniffer for Linux
 * Copyright (C) 2005  Hynek Schlawack <hynek@ularx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* ANSI */
#include <errno.h>
#include <stdio.h> /* only for perror() */
#include <stdlib.h>

/* POSIX */
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

/* libowfat */
#include <buffer.h>
#include <ip4.h>
#include <ip6.h>
#include <str.h>

/* Custom */
#include "dietsniff.h"
#include "rev_res4.h"

/*
** error() - generic complain+quit-function. Depends on errno.
*/
void
error(const char *msg)
{
    int rc = errno;

    if (msg) perror(msg);

    exit(rc);
}

void
print_tcp(struct tcphdr *tcp, struct options *o)
{
    buffer_puts(buffer_1, " [:");
    buffer_putlong(buffer_1, ntohs(tcp->source));
    buffer_puts(buffer_1, " > :");
    buffer_putlong(buffer_1, ntohs(tcp->dest));
    buffer_puts(buffer_1, "]");
}

void
print_udp(struct udphdr *udp, struct options *o)
{
    buffer_puts(buffer_1, " [:");
    buffer_putlong(buffer_1, ntohs(udp->source));
    buffer_puts(buffer_1, " > :");
    buffer_putlong(buffer_1, ntohs(udp->dest));
    buffer_puts(buffer_1, "]");
}

#define ADDR42STR(str, ip) (str[fmt_ip4(str, (char *) &ip)] = '\0', str)

void
print_ip(struct iphdr *ip, struct options *o)
{
    buffer_puts(buffer_1, "[ipv4] ");
    char addr[IP4_FMT] = { 0 };
#ifdef USE_REV_RES
    char *name;

    if (o->no_rev_res) {
#endif /* USE_REV_RES */
	buffer_puts(buffer_1, ADDR42STR(addr, ip->saddr)); 
	buffer_puts(buffer_1, " > ");
	buffer_puts(buffer_1, ADDR42STR(addr, ip->daddr));
#ifdef USE_REV_RES
    } else {
	name = addr2name4(ip->saddr);
	buffer_puts(buffer_1, str_len(name) ? name : ADDR42STR(addr, ip->saddr));
	buffer_puts(buffer_1, " > ");
	name = addr2name4(ip->daddr);
	buffer_puts(buffer_1, str_len(name) ? name : ADDR42STR(addr, ip->daddr));
    }
#endif /* USE_REV_RES */

    /* Handle flags */
    if (o->verbose) {
	buffer_puts(buffer_1, " [");
	if (ip->frag_off) buffer_puts(buffer_1, "DF, ");
	buffer_puts(buffer_1, "LEN=");
	buffer_putlong(buffer_1, ip->tot_len);
	buffer_puts(buffer_1, ", TTL=");
	buffer_putlong(buffer_1, ip->ttl);
	buffer_puts(buffer_1, ", ID=");
	buffer_putlong(buffer_1, ip->id);
	buffer_puts(buffer_1, "]");
    }

    switch(ip->protocol) {
    case 0x1:
	buffer_puts(buffer_1, " [icmpv4]");
	break;
    case 0x2:
	buffer_puts(buffer_1, " [igmpv4]");
	break;
    case 0x6:
	buffer_puts(buffer_1, " [tcp]");
	print_tcp((void *) ip+(ip->ihl * sizeof(long)), o);
	break;
    case 0x11:
	buffer_puts(buffer_1, " [udp]");
	print_udp((void *) ip+(ip->ihl * sizeof(long)), o);
	break;
    default:
	buffer_puts(buffer_1, " [unknown]");
	break;
    }
}

void
print_arp(struct arphdr *arp, struct options *o)
{
    buffer_puts(buffer_1, "[arp] ");
}

void
print_rarp(struct arphdr *arp, struct options *o)
{
    buffer_puts(buffer_1, "[rarp] ");
}

void
print_frame(struct sockaddr_ll *from, char *buf, struct options *o)
{    
    switch (ntohs(from->sll_protocol)) {
	/* IP */
    case ETH_P_IP:
	print_ip((struct iphdr *) buf, o);
	break;
	/* PPPoE */
    case ETH_P_PPP_DISC:
    case ETH_P_PPP_SES:
	buffer_puts(buffer_1, "[PPPoE]");
	break;
	/* ARP */
    case ETH_P_ARP:
	print_arp((struct arphdr *) buf, o);
	break;
	/* RARP */
    case ETH_P_RARP:
	print_arp((struct arphdr *) buf, o);
	break;
    default:
	buffer_puts(buffer_1, "[unknown frametype 0x");
	buffer_putxlong(buffer_1, ntohs(from->sll_protocol));
	buffer_puts(buffer_1, "]");
    }
	
    buffer_putnlflush(buffer_1);
}

/* Caching netdev-name-lookups */
#define MAX_NETDEVS 256 /* Hardcode */
static char *mappings[MAX_NETDEVS];
static inline char *
i2n(int i)
{
    if (__unlikely(!mappings[i])) {
	mappings[i] = malloc(IFNAMSIZ);
	if_indextoname(i, mappings[i]);
    }

    return mappings[i];
}

static inline void
buffer1_put2dec(const unsigned char x) {
    char c[2];
    c[0]=(x/10)+'0';
    c[1]=(x%10)+'0';
    buffer_put(buffer_1,c,2);
}

int
main(int argc, char **argv)
{
    int s;
    char buf[BUF_SIZE];
    struct options o;
    int option;
    struct sockaddr_ll from;
    int fromlen;
    int lo;

    byte_zero(&o, sizeof(struct options));

    while ((option = getopt(argc, argv, "hnvi:")) != -1) {
	switch (option) {
	case 'h':
	    buffer_puts(buffer_1, VER_STRING"\n");
	    buffer_puts(buffer_1, "Usage: dietsniff "
#ifdef USE_REV_RES
			"-n "
#endif /* USE_REV_RES */
			"[-i interface] -v");
	    buffer_putnlflush(buffer_1);

	    exit(0);
	    break;

	case 'i':
	    o.filter.dev = strdup(optarg);
	    break;

#ifdef USE_REV_RES
	case 'n':
	    o.no_rev_res = 1;
	    break;
#endif /* USE_REV_RES */

	case 'v':
	    o.verbose = 1;
	    break;
	}
    }

    if ((s = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ALL))) == -1) error("Creating socket failed");

    fromlen = sizeof(from);

    if (o.filter.dev) {
	from.sll_family = AF_PACKET;
	from.sll_protocol = htons(ETH_P_ALL);
	from.sll_ifindex = if_nametoindex(o.filter.dev);
	if (bind(s, (struct sockaddr *) &from, fromlen) == -1) error("Binding to interface failed");
    }

    /* No error-handling on purpose. If lo doesn't exist (-> -1) we
     * don't leave out packets. */
    lo = if_nametoindex("lo");

    while (1) {
	ssize_t len;
	if ((len = recvfrom(s, buf, BUF_SIZE, 0, (struct sockaddr *) &from, &fromlen)) == -1) error("Receiving frame failed");

	/* Skip duplicate packets on loopback */
	if (from.sll_ifindex == lo && from.sll_pkttype != PACKET_OUTGOING)
	    continue;

	char *name = i2n(from.sll_ifindex);
	struct tm *tm;
	struct timeval tv;

	ioctl(s, SIOCGSTAMP, &tv);
	tm = localtime(&tv.tv_sec);

	buffer1_put2dec(tm->tm_hour);
	buffer_puts(buffer_1, ":");
	buffer1_put2dec(tm->tm_min);
	buffer_puts(buffer_1, ":");
	buffer1_put2dec(tm->tm_sec);
	buffer_puts(buffer_1, ".");
	buffer1_put2dec(tv.tv_usec / 10000);
	buffer1_put2dec((tv.tv_usec / 1000) % 100);
	buffer1_put2dec((tv.tv_usec / 10) % 100);
	buffer_puts(buffer_1, " ");

	buffer_puts(buffer_1, "[");
	buffer_puts(buffer_1, name);
	buffer_puts(buffer_1, "] ");
	print_frame(&from, buf, &o);
    }
	
    return 0;
}
